module MapWeighted(
    Map
  , empty
  , lookup
  , insert
  , delete
  , toList
  ) where

-- Weighted Trees in Haskell

import Prelude hiding(lookup)

-- The weight parameter, says when a tree has become unbalanced.
weight :: Int
weight = 3

-- The data type
data Tree alpha = Null 
            | Node Int (Tree alpha) alpha (Tree alpha)
                              deriving Eq

-- Returns the size of the tree
size :: Tree alpha-> Int
size Null           = 0
size (Node h _ _ _) = h  

-- Constructs a node, maintains size.
-- Pre: arguments are balanced, and size is set correctly
node :: Tree alpha-> alpha-> Tree alpha-> Tree alpha
node l n r = Node h l n r where
               h = 1+ size l+ size r

-- One single rotation to the left.
rotl :: Tree alpha-> Tree alpha
rotl (Node _ xt y (Node _ yt x zt)) = 
  node (node xt y yt) x zt

-- One single rotation to the right.
rotr :: Tree alpha-> Tree alpha
rotr (Node _ (Node _ ut y vt) x rt) = 
  node ut y (node vt x rt)

-- The bias compares the sizes of the left and right subtrees.
bias :: Tree alpha-> Ordering
bias Null = EQ
bias (Node _ lt _ rt) = compare (size lt) (size rt)

-- Main function: construct a balanced tree.
-- Four main cases: 
-- a. Tree small enough 
-- b. Imbalanced to the right 
-- c. Imbalanced to the left
-- d. Balanced
-- Cases b and c require rotations (double/single, depending on bias)
-- Preconditions: trees are balanced, tree to be constructed is
--  "nearly" balanced (ie. just one node too large/small)
mkNode :: Tree alpha-> alpha-> Tree alpha-> Tree alpha
mkNode lt x rt 
  | ls+ rs < 2  = node lt x rt
  | weight* ls < rs  = 
      if bias rt == LT then rotl (node lt x rt)
      else rotl (node lt x (rotr rt))
  | ls > weight* rs =
      if bias lt == GT then rotr (node lt x rt)
      else rotr (node (rotl lt) x rt)
  | otherwise   = node lt x rt where 
    ls = size lt; rs= size rt

{- This is the AVL version:
mkNode xt y zt 
  | sx+1< sz && 
    bias zt < 0 = rotl (node xt y zt)
  | hx+1< hz    = rotl (node xt y (rotr zt))
  | hz+1< hx && 
    bias xt > 0 = rotr (node xt y zt)
  | hz+1< hx    = rotr (node (rotl xt) y zt)
  | otherwise   = node xt y zt
    where hx= ht xt; hz= ht zt
-}

-- Maps built using trees labelled with (key, value) pairs.
-- These functions work on the labelled trees, we wrap them
-- into functions on the map data type below.

lookup' :: Ord alpha=> alpha-> Tree (alpha, beta)-> Maybe beta
lookup' k Null = Nothing
lookup' k (Node _ lt (kv, kn) rt) 
  | k == kv  = Just kn
  | k <  kv  = lookup' k lt
  | k >  kv  = lookup' k rt

insert' :: Ord alpha=> alpha-> beta-> Tree (alpha, beta)-> Tree (alpha, beta)
insert' k v Null = node Null (k, v) Null
insert' k v (Node n l a@(kn, _) r) 
  | k < kn  = mkNode (insert' k v l) a r
  | k == kn = Node n l (k, v) r
  | k > kn  = mkNode l a (insert' k v r)

delete' :: Ord alpha=> alpha-> Tree (alpha, beta)-> Tree (alpha, beta)
delete' k Null = Null
delete' k (Node s l a@(kn, _) r)
  | k < kn  = mkNode (delete' k l) a r
  | k == kn = join l r
  | k > kn  = mkNode l a (delete' k r)

-- Joins two trees
-- Preconditions: 
--   |size l- size r| <= 1
--   for all node values n in l und m in r, n < m.
join ::  Tree alpha-> Tree alpha-> Tree alpha
join xt Null = xt
join lt rt   = mkNode lt u nu where
  (u, nu) = splitTree rt
  splitTree :: Tree alpha-> (alpha, Tree alpha)
  splitTree (Node h Null a t) = (a, t)
  splitTree (Node h lt a rt) =
    (u, mkNode nu a rt) where
       (u, nu) = splitTree lt

fold :: (alpha-> beta-> beta)-> beta-> Tree alpha-> beta    
fold f e Null = e
fold f e (Node _ l a r) =
   fold f (f a (fold f e r)) l

-- Slightly fancy Show instance
instance Show alpha=> Show (Tree alpha) where
  show t = shw 0 t where
    shw _ Null = []
    shw n (Node _ l a r) = (shw (n+1) l)
                              ++ spc n ++ show a  ++"\n" 
                              ++ (shw (n+1) r)
    spc n = concat (replicate n "   ")

-- | Maps are trees labelled with pairs of (key, value),
--   and ordered by the keys.

data Map alpha beta = Map { tree :: Tree (alpha, beta) }
                      deriving (Eq, Show)

empty :: Map alpha beta
empty = Map Null

lookup ::  Ord alpha=> alpha-> Map alpha beta-> Maybe beta
lookup k (Map t) = lookup' k t

insert :: Ord alpha=> alpha-> beta-> Map alpha beta-> Map alpha beta
insert k v (Map t) = Map (insert' k v t)

delete :: Ord alpha=> alpha-> Map alpha beta-> Map alpha beta
delete k (Map t) = Map (delete' k t)

toList :: Map alpha beta -> [(alpha, beta)]
toList (Map t) = fold (:) [] t



-- Test trees
t :: [Int]-> Map Int String
t = foldl (\t n -> insert n (show n) t) empty

test :: [Int]-> [Map Int String]
test = scanl (\t n -> insert n (show n) t) empty

t1 = t [4,5,7,2,1,3,6,8,9]
t2 = t [3,1,4,3]



